@@ -1,14 +1,15 @@
 /* =========================================================================
-   Copyright (c) 2010-2011, Institute for Microelectronics,
+   Copyright (c) 2010-2014, Institute for Microelectronics,
                             Institute for Analysis and Scientific Computing,
                             TU Wien.
+   Portions of this software are copyright by UChicago Argonne, LLC.
 
                             -----------------
                   ViennaCL - The Vienna Computing Library
                             -----------------
 
    Project Head:    Karl Rupp                   rupp@iue.tuwien.ac.at
-               
+
    (A list of authors and contributors can be found in the PDF manual)
 
    License:         MIT (X11), see file LICENSE in the base directory
@@ -22,6 +23,8 @@
 
 //#define VIENNACL_BUILD_INFO
 
+//#define VIENNACL_DEBUG_ALL
+
 #include "viennacl/toeplitz_matrix.hpp"
 #include "viennacl/circulant_matrix.hpp"
 #include "viennacl/vandermonde_matrix.hpp"
@@ -37,18 +40,20 @@ template <typename T>
 class dense_matrix
 {
   public:
+    typedef std::size_t   size_type;
+
     dense_matrix(std::size_t rows, std::size_t cols) : elements_(rows * cols), rows_(rows), cols_(cols) {}
-    
+
     T & operator()(std::size_t i, std::size_t j) { return elements_[i*cols_ + j]; }
     T const & operator()(std::size_t i, std::size_t j) const { return elements_[i*cols_ + j]; }
-    
+
     std::size_t size1() const { return rows_; }
     std::size_t size2() const { return cols_; }
 
     dense_matrix & operator+=(dense_matrix const & other)
     {
-      for(std::size_t i = 0; i < other.size1(); i++) 
-        for(std::size_t j = 0; j < other.size2(); j++) 
+      for(std::size_t i = 0; i < other.size1(); i++)
+        for(std::size_t j = 0; j < other.size2(); j++)
           elements_[i*cols_ + j] = other.elements_[i*cols_+j];
       return *this;
     }
@@ -70,20 +75,20 @@ std::ostream & operator<<(std::ostream & os, dense_matrix<T> const & mat)
       std::cout << mat(i,j) << ",";
     std::cout << ")";
   }
-  
+
   return os;
 }
 
 
 template <typename ScalarType>
-ScalarType diff(dense_matrix<ScalarType> const & m1, dense_matrix<ScalarType> const & m2) 
+ScalarType diff(dense_matrix<ScalarType> const & m1, dense_matrix<ScalarType> const & m2)
 {
     ScalarType df = 0.0;
     ScalarType d1 = 0;
     ScalarType d2 = 0;
 
-    for(std::size_t i = 0; i < m1.size1(); i++) 
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+    for(std::size_t i = 0; i < m1.size1(); i++)
+      for(std::size_t j = 0; j < m1.size2(); j++)
       {
         df += (m1(i,j) - m2(i,j)) * (m1(i,j) - m2(i,j));
         d1 += m1(i,j) * m1(i,j);
@@ -92,45 +97,45 @@ ScalarType diff(dense_matrix<ScalarType> const & m1, dense_matrix<ScalarType> co
 
     if ( (d1 == 0) && (d2 == 0) )
       return 0;
-    
-    return sqrt(df / fmax(d1, d2));
+
+    return std::sqrt(df / std::max<ScalarType>(d1, d2));
 }
 
 
 template <typename ScalarType>
-ScalarType diff(std::vector<ScalarType>& vec, std::vector<ScalarType>& ref) 
+ScalarType diff(std::vector<ScalarType>& vec, std::vector<ScalarType>& ref)
 {
     ScalarType df = 0.0;
     ScalarType norm_ref = 0;
 
-    for(std::size_t i = 0; i < vec.size(); i++) 
+    for(std::size_t i = 0; i < vec.size(); i++)
     {
         df = df + pow(vec[i] - ref[i], 2);
         norm_ref += ref[i] * ref[i];
     }
 
-    return sqrt(df / norm_ref) ;
+    return std::sqrt(df / norm_ref) ;
 }
 
 template <typename ScalarType>
-ScalarType diff_max(std::vector<ScalarType>& vec, std::vector<ScalarType>& ref) 
+ScalarType diff_max(std::vector<ScalarType>& vec, std::vector<ScalarType>& ref)
 {
   ScalarType df = 0.0;
   ScalarType mx = 0.0;
   ScalarType norm_max = 0;
-  
-  for (std::size_t i = 0; i < vec.size(); i++) 
+
+  for (std::size_t i = 0; i < vec.size(); i++)
   {
-    df = fmax(fabs(vec[i] - ref[i]), df);
-    mx = fmax(fabs(vec[i]), mx);
-    
+    df = std::max<ScalarType>(fabs(vec[i] - ref[i]), df);
+    mx = std::max<ScalarType>(fabs(vec[i]), mx);
+
     if (mx > 0)
     {
       if (norm_max < df / mx)
         norm_max = df / mx;
     }
   }
-  
+
   return norm_max;
 }
 
@@ -156,7 +161,7 @@ void transpose_test()
 
 
 template <typename ScalarType>
-int toeplitz_test(ScalarType epsilon) 
+int toeplitz_test(ScalarType epsilon)
 {
     std::size_t TOEPLITZ_SIZE = 47;
     viennacl::toeplitz_matrix<ScalarType> vcl_toeplitz1(TOEPLITZ_SIZE, TOEPLITZ_SIZE);
@@ -167,15 +172,15 @@ int toeplitz_test(ScalarType epsilon)
 
     std::vector<ScalarType> input_ref(TOEPLITZ_SIZE);
     std::vector<ScalarType> result_ref(TOEPLITZ_SIZE);
-    
+
     dense_matrix<ScalarType> m1(TOEPLITZ_SIZE, TOEPLITZ_SIZE);
     dense_matrix<ScalarType> m2(TOEPLITZ_SIZE, TOEPLITZ_SIZE);
 
-    for(std::size_t i = 0; i < TOEPLITZ_SIZE; i++) 
-      for(std::size_t j = 0; j < TOEPLITZ_SIZE; j++) 
+    for(std::size_t i = 0; i < TOEPLITZ_SIZE; i++)
+      for(std::size_t j = 0; j < TOEPLITZ_SIZE; j++)
       {
         m1(i,j) = static_cast<ScalarType>(i) - static_cast<ScalarType>(j);
-        m2(i,j) = m1(i,j) * m1(i,j) + 1;
+        m2(i,j) = m1(i,j) * m1(i,j) + ScalarType(1);
       }
 
     for(std::size_t i = 0; i < TOEPLITZ_SIZE; i++)
@@ -185,21 +190,21 @@ int toeplitz_test(ScalarType epsilon)
     viennacl::copy(m1, vcl_toeplitz1);
     viennacl::copy(m2, vcl_toeplitz2);
     viennacl::copy(input_ref, vcl_input);
-    
+
     //
     // Matrix-Vector product:
     //
     vcl_result = viennacl::linalg::prod(vcl_toeplitz1, vcl_input);
-    
+
     for(std::size_t i = 0; i < m1.size1(); i++)     //reference calculation
     {
       ScalarType entry = 0;
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+      for(std::size_t j = 0; j < m1.size2(); j++)
         entry += m1(i,j) * input_ref[j];
-      
+
       result_ref[i] = entry;
     }
-    
+
     viennacl::copy(vcl_result, input_ref);
     std::cout << "Matrix-Vector Product: " << diff_max(input_ref, result_ref);
     if (diff_max(input_ref, result_ref) < epsilon)
@@ -211,17 +216,17 @@ int toeplitz_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
-    
+
+
     //
     // Matrix addition:
     //
     vcl_toeplitz1 += vcl_toeplitz2;
 
     for(std::size_t i = 0; i < m1.size1(); i++)    //reference calculation
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+      for(std::size_t j = 0; j < m1.size2(); j++)
         m1(i,j) += m2(i,j);
-    
+
     viennacl::copy(vcl_toeplitz1, m2);
     std::cout << "Matrix Addition: " << diff(m1, m2);
     if (diff(m1, m2) < epsilon)
@@ -231,12 +236,12 @@ int toeplitz_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
+
     //
     // Per-Element access:
     //
     vcl_toeplitz1(2,4) = 42;
-    
+
     for (std::size_t i=0; i<m1.size1(); ++i)    //reference calculation
     {
       if (i + 2 < m1.size2())
@@ -252,12 +257,12 @@ int toeplitz_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
+
     return EXIT_SUCCESS;
 }
 
 template <typename ScalarType>
-int circulant_test(ScalarType epsilon) 
+int circulant_test(ScalarType epsilon)
 {
     std::size_t CIRCULANT_SIZE = 53;
     viennacl::circulant_matrix<ScalarType> vcl_circulant1(CIRCULANT_SIZE, CIRCULANT_SIZE);
@@ -268,12 +273,12 @@ int circulant_test(ScalarType epsilon)
 
     std::vector<ScalarType> input_ref(CIRCULANT_SIZE);
     std::vector<ScalarType> result_ref(CIRCULANT_SIZE);
-    
+
     dense_matrix<ScalarType> m1(vcl_circulant1.size1(), vcl_circulant1.size2());
     dense_matrix<ScalarType> m2(vcl_circulant1.size1(), vcl_circulant1.size2());
 
-    for(std::size_t i = 0; i < m1.size1(); i++) 
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+    for(std::size_t i = 0; i < m1.size1(); i++)
+      for(std::size_t j = 0; j < m1.size2(); j++)
       {
         m1(i,j) = static_cast<ScalarType>((i - j + m1.size1()) % m1.size1());
         m2(i,j) = m1(i,j) * m1(i,j) + ScalarType(1);
@@ -281,26 +286,26 @@ int circulant_test(ScalarType epsilon)
 
     for(std::size_t i = 0; i < input_ref.size(); i++)
       input_ref[i] = ScalarType(i);
-    
+
     // Copy to ViennaCL
     viennacl::copy(m1, vcl_circulant1);
     viennacl::copy(m2, vcl_circulant2);
     viennacl::copy(input_ref, vcl_input);
-    
+
     //
     // Matrix-Vector product:
     //
     vcl_result = viennacl::linalg::prod(vcl_circulant1, vcl_input);
-    
+
     for(std::size_t i = 0; i < m1.size1(); i++)     //reference calculation
     {
       ScalarType entry = 0;
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+      for(std::size_t j = 0; j < m1.size2(); j++)
         entry += m1(i,j) * input_ref[j];
-      
+
       result_ref[i] = entry;
     }
-    
+
     viennacl::copy(vcl_result, input_ref);
     std::cout << "Matrix-Vector Product: " << diff_max(input_ref, result_ref);
     if (diff_max(input_ref, result_ref) < epsilon)
@@ -312,17 +317,17 @@ int circulant_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
-    
+
+
     //
     // Matrix addition:
     //
     vcl_circulant1 += vcl_circulant2;
 
     for(std::size_t i = 0; i < m1.size1(); i++)    //reference calculation
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+      for(std::size_t j = 0; j < m1.size2(); j++)
         m1(i,j) += m2(i,j);
-    
+
     viennacl::copy(vcl_circulant1, m2);
     std::cout << "Matrix Addition: " << diff(m1, m2);
     if (diff(m1, m2) < epsilon)
@@ -332,14 +337,14 @@ int circulant_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
+
     //
     // Per-Element access:
     //
     vcl_circulant1(4,2) = 42;
-    
+
     for(std::size_t i = 0; i < m1.size1(); i++)    //reference calculation
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+      for(std::size_t j = 0; j < m1.size2(); j++)
       {
         if ((i - j + m1.size1()) % m1.size1() == 2)
           m1(i, j) = 42;
@@ -354,12 +359,12 @@ int circulant_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
+
     return EXIT_SUCCESS;
 }
 
 template <typename ScalarType>
-int vandermonde_test(ScalarType epsilon) 
+int vandermonde_test(ScalarType epsilon)
 {
     std::size_t VANDERMONDE_SIZE = 61;
 
@@ -371,39 +376,39 @@ int vandermonde_test(ScalarType epsilon)
 
     std::vector<ScalarType> input_ref(VANDERMONDE_SIZE);
     std::vector<ScalarType> result_ref(VANDERMONDE_SIZE);
-    
+
     dense_matrix<ScalarType> m1(vcl_vandermonde1.size1(), vcl_vandermonde1.size2());
     dense_matrix<ScalarType> m2(m1.size1(), m1.size2());
 
-    for(std::size_t i = 0; i < m1.size1(); i++) 
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+    for(std::size_t i = 0; i < m1.size1(); i++)
+      for(std::size_t j = 0; j < m1.size2(); j++)
       {
-        m1(i,j) = pow(ScalarType(1.0 + i/1000.0), ScalarType(j));
-        m2(i,j) = pow(ScalarType(1.0 - i/2000.0), ScalarType(j));
+        m1(i,j) = std::pow(ScalarType(1.0 + i/1000.0), ScalarType(j));
+        m2(i,j) = std::pow(ScalarType(1.0 - i/2000.0), ScalarType(j));
       }
 
     for(std::size_t i = 0; i < input_ref.size(); i++)
       input_ref[i] = ScalarType(i);
-    
+
     // Copy to ViennaCL
     viennacl::copy(m1, vcl_vandermonde1);
     viennacl::copy(m2, vcl_vandermonde2);
     viennacl::copy(input_ref, vcl_input);
-    
+
     //
     // Matrix-Vector product:
     //
     vcl_result = viennacl::linalg::prod(vcl_vandermonde1, vcl_input);
-    
+
     for(std::size_t i = 0; i < m1.size1(); i++)     //reference calculation
     {
       ScalarType entry = 0;
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+      for(std::size_t j = 0; j < m1.size2(); j++)
         entry += m1(i,j) * input_ref[j];
-      
+
       result_ref[i] = entry;
     }
-    
+
     viennacl::copy(vcl_result, input_ref);
     std::cout << "Matrix-Vector Product: " << diff_max(input_ref, result_ref);
     if (diff_max(input_ref, result_ref) < epsilon)
@@ -415,21 +420,21 @@ int vandermonde_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
-    
+
+
     //
     // Note: Matrix addition does not make sense for a Vandermonde matrix
     //
 
-    
+
     //
     // Per-Element access:
     //
-    vcl_vandermonde1(4) = 1.0001;
-    
-    for(std::size_t j = 0; j < m1.size2(); j++) 
+    vcl_vandermonde1(4) = static_cast<ScalarType>(1.0001);
+
+    for(std::size_t j = 0; j < m1.size2(); j++)
     {
-      m1(4, j) = pow(ScalarType(1.0001), ScalarType(j));
+      m1(4, j) = std::pow(ScalarType(1.0001), ScalarType(j));
     }
 
     viennacl::copy(vcl_vandermonde1, m2);
@@ -441,7 +446,7 @@ int vandermonde_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
+
     return EXIT_SUCCESS;
 }
 
@@ -457,12 +462,12 @@ int hankel_test(ScalarType epsilon)
 
     std::vector<ScalarType> input_ref(HANKEL_SIZE);
     std::vector<ScalarType> result_ref(HANKEL_SIZE);
-    
+
     dense_matrix<ScalarType> m1(vcl_hankel1.size1(), vcl_hankel1.size2());
     dense_matrix<ScalarType> m2(m1.size1(), m1.size2());
 
-    for(std::size_t i = 0; i < m1.size1(); i++) 
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+    for(std::size_t i = 0; i < m1.size1(); i++)
+      for(std::size_t j = 0; j < m1.size2(); j++)
       {
         m1(i,j) = static_cast<ScalarType>((i + j) % (2 * m1.size1()));
         m2(i,j) = m1(i,j) * m1(i,j) + ScalarType(1);
@@ -470,26 +475,26 @@ int hankel_test(ScalarType epsilon)
 
     for(std::size_t i = 0; i < input_ref.size(); i++)
       input_ref[i] = ScalarType(i);
-    
+
     // Copy to ViennaCL
     viennacl::copy(m1, vcl_hankel1);
     viennacl::copy(m2, vcl_hankel2);
     viennacl::copy(input_ref, vcl_input);
-    
+
     //
     // Matrix-Vector product:
     //
     vcl_result = viennacl::linalg::prod(vcl_hankel1, vcl_input);
-    
+
     for(std::size_t i = 0; i < m1.size1(); i++)     //reference calculation
     {
       ScalarType entry = 0;
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+      for(std::size_t j = 0; j < m1.size2(); j++)
         entry += m1(i,j) * input_ref[j];
-      
+
       result_ref[i] = entry;
     }
-    
+
     viennacl::copy(vcl_result, input_ref);
     std::cout << "Matrix-Vector Product: " << diff_max(input_ref, result_ref);
     if (diff_max(input_ref, result_ref) < epsilon)
@@ -501,17 +506,17 @@ int hankel_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
-    
+
+
     //
     // Matrix addition:
     //
     vcl_hankel1 += vcl_hankel2;
 
     for(std::size_t i = 0; i < m1.size1(); i++)    //reference calculation
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+      for(std::size_t j = 0; j < m1.size2(); j++)
         m1(i,j) += m2(i,j);
-    
+
     viennacl::copy(vcl_hankel1, m2);
     std::cout << "Matrix Addition: " << diff(m1, m2);
     if (diff(m1, m2) < epsilon)
@@ -521,14 +526,14 @@ int hankel_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
+
     //
     // Per-Element access:
     //
     vcl_hankel1(4,2) = 42;
-    
+
     for(std::size_t i = 0; i < m1.size1(); i++)    //reference calculation
-      for(std::size_t j = 0; j < m1.size2(); j++) 
+      for(std::size_t j = 0; j < m1.size2(); j++)
       {
         if ((i + j) % (2*m1.size1()) == 6)
           m1(i, j) = 42;
@@ -543,11 +548,11 @@ int hankel_test(ScalarType epsilon)
       std::cout << " [FAILED]" << std::endl;
       return EXIT_FAILURE;
     }
-    
+
     return EXIT_SUCCESS;
 }
 
-int main() 
+int main()
 {
   std::cout << std::endl;
   std::cout << "----------------------------------------------" << std::endl;
@@ -558,7 +563,7 @@ int main()
   std::cout << std::endl;
 
   double eps = 1e-3;
-  
+
   std::cout << "# Testing setup:" << std::endl;
   std::cout << "  eps:     " << eps << std::endl;
   std::cout << "  numeric: float" << std::endl;
@@ -566,48 +571,53 @@ int main()
   std::cout << " -- Vandermonde matrix -- " << std::endl;
   if (vandermonde_test<float>(static_cast<float>(eps)) == EXIT_FAILURE)
     return EXIT_FAILURE;
-  
+
   std::cout << " -- Circulant matrix -- " << std::endl;
   if (circulant_test<float>(static_cast<float>(eps)) == EXIT_FAILURE)
     return EXIT_FAILURE;
-  
+
   std::cout << " -- Toeplitz matrix -- " << std::endl;
   if (toeplitz_test<float>(static_cast<float>(eps)) == EXIT_FAILURE)
     return EXIT_FAILURE;
-  
+
   std::cout << " -- Hankel matrix -- " << std::endl;
   if (hankel_test<float>(static_cast<float>(eps)) == EXIT_FAILURE)
     return EXIT_FAILURE;
-  
-  
+
+
   std::cout << std::endl;
 
   if( viennacl::ocl::current_device().double_support() )
   {
     eps = 1e-10;
-    
+
     std::cout << std::endl;
     std::cout << "# Testing setup:" << std::endl;
     std::cout << "  eps:     " << eps << std::endl;
     std::cout << "  numeric: double" << std::endl;
     std::cout << std::endl;
-    
+
     std::cout << " -- Vandermonde matrix -- " << std::endl;
     if (vandermonde_test<double>(eps) == EXIT_FAILURE)
       return EXIT_FAILURE;
-    
+
     std::cout << " -- Circulant matrix -- " << std::endl;
     if (circulant_test<double>(eps) == EXIT_FAILURE)
       return EXIT_FAILURE;
-    
+
     std::cout << " -- Toeplitz matrix -- " << std::endl;
     if (toeplitz_test<double>(eps) == EXIT_FAILURE)
       return EXIT_FAILURE;
-    
+
     std::cout << " -- Hankel matrix -- " << std::endl;
     if (hankel_test<double>(eps) == EXIT_FAILURE)
       return EXIT_FAILURE;
   }
 
+  std::cout << std::endl;
+  std::cout << "------- Test completed --------" << std::endl;
+  std::cout << std::endl;
+
+
   return EXIT_SUCCESS;
 }
