@@ -2,41 +2,42 @@
 #define VIENNACL_FFT_HPP
 
 /* =========================================================================
-   Copyright (c) 2010-2011, Institute for Microelectronics,
+   Copyright (c) 2010-2014, Institute for Microelectronics,
                             Institute for Analysis and Scientific Computing,
                             TU Wien.
+   Portions of this software are copyright by UChicago Argonne, LLC.
 
                             -----------------
                   ViennaCL - The Vienna Computing Library
                             -----------------
 
    Project Head:    Karl Rupp                   rupp@iue.tuwien.ac.at
-               
+
    (A list of authors and contributors can be found in the PDF manual)
 
    License:         MIT (X11), see file LICENSE in the base directory
 ============================================================================= */
 
-/** @file fft.hpp
-    @brief All routines related to the Fast Fourier Transform. Experimental in 1.2.x.
+/** @file viennacl/fft.hpp
+    @brief All routines related to the Fast Fourier Transform. Experimental.
 */
 
 #include <viennacl/vector.hpp>
 #include <viennacl/matrix.hpp>
 
-#include "viennacl/linalg/kernels/fft_kernels.h"
+#include "viennacl/linalg/opencl/kernels/fft.hpp"
 
 #include <cmath>
 
 #include <stdexcept>
 
-namespace viennacl 
+namespace viennacl
 {
   namespace detail
   {
     namespace fft
     {
-        const std::size_t MAX_LOCAL_POINTS_NUM = 512;
+        const vcl_size_t MAX_LOCAL_POINTS_NUM = 512;
 
         namespace FFT_DATA_ORDER {
             enum DATA_ORDER {
@@ -49,21 +50,24 @@ namespace viennacl
 }
 
 /// @cond
-namespace viennacl {
-  namespace detail {
-    namespace fft {
+namespace viennacl
+{
+  namespace detail
+  {
+    namespace fft
+    {
 
-        inline bool is_radix2(std::size_t data_size) {
+        inline bool is_radix2(vcl_size_t data_size) {
             return !((data_size > 2) && (data_size & (data_size - 1)));
 
         }
 
-        inline std::size_t next_power_2(std::size_t n) {
+        inline vcl_size_t next_power_2(vcl_size_t n) {
             n = n - 1;
 
-            std::size_t power = 1;
+            vcl_size_t power = 1;
 
-            while(power < sizeof(std::size_t) * 8) {
+            while(power < sizeof(vcl_size_t) * 8) {
                 n = n | (n >> power);
                 power *= 2;
             }
@@ -71,10 +75,10 @@ namespace viennacl {
             return n + 1;
         }
 
-        inline std::size_t num_bits(std::size_t size)
+        inline vcl_size_t num_bits(vcl_size_t size)
         {
-            std::size_t bits_datasize = 0;
-            std::size_t ds = 1;
+            vcl_size_t bits_datasize = 0;
+            vcl_size_t ds = 1;
 
             while(ds < size)
             {
@@ -95,21 +99,25 @@ namespace viennacl {
         template<class SCALARTYPE>
         void direct(const viennacl::ocl::handle<cl_mem>& in,
                     const viennacl::ocl::handle<cl_mem>& out,
-                    std::size_t size,
-                    std::size_t stride,
-                    std::size_t batch_num,
+                    vcl_size_t size,
+                    vcl_size_t stride,
+                    vcl_size_t batch_num,
                     SCALARTYPE sign = -1.0f,
                     FFT_DATA_ORDER::DATA_ORDER data_order = FFT_DATA_ORDER::ROW_MAJOR
                     )
         {
-          viennacl::linalg::kernels::matrix_row<SCALARTYPE, 1>::init();
-          std::string program_string = viennacl::linalg::kernels::matrix_row<SCALARTYPE, 1>::program_name();
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(in.context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
+
+          std::string program_string = viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, row_major>::program_name();
           if (data_order == FFT_DATA_ORDER::COL_MAJOR)
           {
-            viennacl::linalg::kernels::matrix_col<SCALARTYPE, 1>::init();
-            program_string = viennacl::linalg::kernels::matrix_col<SCALARTYPE, 1>::program_name();
+            viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, column_major>::init(ctx);
+            program_string = viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, column_major>::program_name();
           }
-          viennacl::ocl::kernel& kernel = viennacl::ocl::current_context().get_program(program_string).get_kernel("fft_direct");
+          else
+            viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, row_major>::init(ctx);
+          viennacl::ocl::kernel& kernel = ctx.get_kernel(program_string, "fft_direct");
           viennacl::ocl::enqueue(kernel(in, out, static_cast<cl_uint>(size), static_cast<cl_uint>(stride), static_cast<cl_uint>(batch_num), sign));
         }
 
@@ -119,25 +127,27 @@ namespace viennacl {
         */
         template <typename SCALARTYPE>
         void reorder(const viennacl::ocl::handle<cl_mem>& in,
-                     std::size_t size,
-                     std::size_t stride,
-                     std::size_t bits_datasize,
-                     std::size_t batch_num,
+                     vcl_size_t size,
+                     vcl_size_t stride,
+                     vcl_size_t bits_datasize,
+                     vcl_size_t batch_num,
                      FFT_DATA_ORDER::DATA_ORDER data_order = FFT_DATA_ORDER::ROW_MAJOR
                      )
         {
-          viennacl::linalg::kernels::matrix_row<SCALARTYPE, 1>::init();
-          std::string program_string = viennacl::linalg::kernels::matrix_row<SCALARTYPE, 1>::program_name();
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(in.context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
+
+          std::string program_string = viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, row_major>::program_name();
           if (data_order == FFT_DATA_ORDER::COL_MAJOR)
           {
-            viennacl::linalg::kernels::matrix_col<SCALARTYPE, 1>::init();
-            program_string = viennacl::linalg::kernels::matrix_col<SCALARTYPE, 1>::program_name();
+            viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, column_major>::init(ctx);
+            program_string = viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, column_major>::program_name();
           }
-          
-          viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                              .get_program(program_string)
-                                              .get_kernel("fft_reorder");
-          viennacl::ocl::enqueue(kernel(in, 
+          else
+            viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, row_major>::init(ctx);
+
+          viennacl::ocl::kernel& kernel = ctx.get_kernel(program_string, "fft_reorder");
+          viennacl::ocl::enqueue(kernel(in,
                                         static_cast<cl_uint>(bits_datasize),
                                         static_cast<cl_uint>(size),
                                         static_cast<cl_uint>(stride),
@@ -155,33 +165,33 @@ namespace viennacl {
         */
         template<class SCALARTYPE>
         void radix2(const viennacl::ocl::handle<cl_mem>& in,
-                    std::size_t size,
-                    std::size_t stride,
-                    std::size_t batch_num,
+                    vcl_size_t size,
+                    vcl_size_t stride,
+                    vcl_size_t batch_num,
                     SCALARTYPE sign = -1.0f,
                     FFT_DATA_ORDER::DATA_ORDER data_order = FFT_DATA_ORDER::ROW_MAJOR
                     )
         {
-          viennacl::linalg::kernels::fft<SCALARTYPE, 1>::init();
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(in.context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
 
             assert(batch_num != 0);
             assert(is_radix2(size));
 
-            viennacl::linalg::kernels::matrix_row<SCALARTYPE, 1>::init();
-            std::string program_string = viennacl::linalg::kernels::matrix_row<SCALARTYPE, 1>::program_name();
+            std::string program_string = viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, row_major>::program_name();
             if (data_order == FFT_DATA_ORDER::COL_MAJOR)
             {
-              viennacl::linalg::kernels::matrix_col<SCALARTYPE, 1>::init();
-              program_string = viennacl::linalg::kernels::matrix_col<SCALARTYPE, 1>::program_name();
+              viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, column_major>::init(ctx);
+              program_string = viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, column_major>::program_name();
             }
+            else
+              viennacl::linalg::opencl::kernels::matrix<SCALARTYPE, row_major>::init(ctx);
 
-            std::size_t bits_datasize = num_bits(size);
+            vcl_size_t bits_datasize = num_bits(size);
 
-            if(size <= MAX_LOCAL_POINTS_NUM) 
+            if(size <= MAX_LOCAL_POINTS_NUM)
             {
-                viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                                 .get_program(program_string)
-                                                 .get_kernel("fft_radix2_local");
+                viennacl::ocl::kernel& kernel = ctx.get_kernel(program_string, "fft_radix2_local");
                 viennacl::ocl::enqueue(kernel(in,
                                               viennacl::ocl::local_mem((size * 4) * sizeof(SCALARTYPE)),
                                               static_cast<cl_uint>(bits_datasize),
@@ -189,16 +199,14 @@ namespace viennacl {
                                               static_cast<cl_uint>(stride),
                                               static_cast<cl_uint>(batch_num),
                                               sign));
-            } 
+            }
             else
             {
                 reorder<SCALARTYPE>(in, size, stride, bits_datasize, batch_num);
 
-                for(std::size_t step = 0; step < bits_datasize; step++) 
+                for(vcl_size_t step = 0; step < bits_datasize; step++)
                 {
-                    viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                                     .get_program(program_string)
-                                                     .get_kernel("fft_radix2");
+                    viennacl::ocl::kernel& kernel = ctx.get_kernel(program_string, "fft_radix2");
                     viennacl::ocl::enqueue(kernel(in,
                                                   static_cast<cl_uint>(step),
                                                   static_cast<cl_uint>(bits_datasize),
@@ -221,24 +229,21 @@ namespace viennacl {
         template<class SCALARTYPE, unsigned int ALIGNMENT>
         void bluestein(viennacl::vector<SCALARTYPE, ALIGNMENT>& in,
                        viennacl::vector<SCALARTYPE, ALIGNMENT>& out,
-                       std::size_t batch_num,
-                       SCALARTYPE sign = -1.0
-                       )
+                       vcl_size_t /*batch_num*/)
         {
-          viennacl::linalg::kernels::fft<SCALARTYPE, 1>::init();
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(viennacl::traits::opencl_handle(in).context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
 
-            std::size_t size = in.size() >> 1;
-            std::size_t ext_size = next_power_2(2 * size - 1);
+          vcl_size_t size = in.size() >> 1;
+          vcl_size_t ext_size = next_power_2(2 * size - 1);
 
-            viennacl::vector<SCALARTYPE, ALIGNMENT> A(ext_size << 1);
-            viennacl::vector<SCALARTYPE, ALIGNMENT> B(ext_size << 1);
+          viennacl::vector<SCALARTYPE, ALIGNMENT> A(ext_size << 1);
+          viennacl::vector<SCALARTYPE, ALIGNMENT> B(ext_size << 1);
 
-            viennacl::vector<SCALARTYPE, ALIGNMENT> Z(ext_size << 1);
+          viennacl::vector<SCALARTYPE, ALIGNMENT> Z(ext_size << 1);
 
             {
-                viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                             .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                             .get_kernel("zero2");
+                viennacl::ocl::kernel& kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "zero2");
                 viennacl::ocl::enqueue(kernel(
                                             A,
                                             B,
@@ -247,9 +252,7 @@ namespace viennacl {
 
             }
             {
-                viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                             .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                             .get_kernel("bluestein_pre");
+                viennacl::ocl::kernel& kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "bluestein_pre");
                 viennacl::ocl::enqueue(kernel(
                                            in,
                                            A,
@@ -262,9 +265,7 @@ namespace viennacl {
             viennacl::linalg::convolve_i(A, B, Z);
 
             {
-                viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                                 .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                                 .get_kernel("bluestein_post");
+                viennacl::ocl::kernel& kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "bluestein_post");
                 viennacl::ocl::enqueue(kernel(
                                             Z,
                                             out,
@@ -276,95 +277,90 @@ namespace viennacl {
         template<class SCALARTYPE, unsigned int ALIGNMENT>
         void multiply(viennacl::vector<SCALARTYPE, ALIGNMENT> const & input1,
                       viennacl::vector<SCALARTYPE, ALIGNMENT> const & input2,
-                      viennacl::vector<SCALARTYPE, ALIGNMENT> & output) 
+                      viennacl::vector<SCALARTYPE, ALIGNMENT> & output)
         {
-          viennacl::linalg::kernels::fft<SCALARTYPE, 1>::init();
-            std::size_t size = input1.size() >> 1;
-            viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                             .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                             .get_kernel("fft_mult_vec");
-            viennacl::ocl::enqueue(kernel(input1, input2, output, static_cast<cl_uint>(size)));
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(viennacl::traits::opencl_handle(input1).context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
+          vcl_size_t size = input1.size() >> 1;
+          viennacl::ocl::kernel& kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "fft_mult_vec");
+          viennacl::ocl::enqueue(kernel(input1, input2, output, static_cast<cl_uint>(size)));
         }
 
         template<class SCALARTYPE, unsigned int ALIGNMENT>
-        void normalize(viennacl::vector<SCALARTYPE, ALIGNMENT> & input) 
+        void normalize(viennacl::vector<SCALARTYPE, ALIGNMENT> & input)
         {
-          viennacl::linalg::kernels::fft<SCALARTYPE, 1>::init();
-            viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                             .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                             .get_kernel("fft_div_vec_scalar");
-            std::size_t size = input.size() >> 1;
-            SCALARTYPE norm_factor = static_cast<SCALARTYPE>(size);
-            viennacl::ocl::enqueue(kernel(input, static_cast<cl_uint>(size), norm_factor));
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(viennacl::traits::opencl_handle(input).context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
+
+          viennacl::ocl::kernel& kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "fft_div_vec_scalar");
+          vcl_size_t size = input.size() >> 1;
+          SCALARTYPE norm_factor = static_cast<SCALARTYPE>(size);
+          viennacl::ocl::enqueue(kernel(input, static_cast<cl_uint>(size), norm_factor));
         }
 
         template<class SCALARTYPE, unsigned int ALIGNMENT>
-        void transpose(viennacl::matrix<SCALARTYPE, viennacl::row_major, ALIGNMENT> & input) 
+        void transpose(viennacl::matrix<SCALARTYPE, viennacl::row_major, ALIGNMENT> & input)
         {
-          viennacl::linalg::kernels::fft<SCALARTYPE, 1>::init();
-            viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                             .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                             .get_kernel("transpose_inplace");
-            viennacl::ocl::enqueue(kernel(input,
-                                          static_cast<cl_uint>(input.internal_size1()),
-                                          static_cast<cl_uint>(input.internal_size2()) >> 1));
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(viennacl::traits::opencl_handle(input).context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
+
+          viennacl::ocl::kernel& kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "transpose_inplace");
+          viennacl::ocl::enqueue(kernel(input,
+                                        static_cast<cl_uint>(input.internal_size1()),
+                                        static_cast<cl_uint>(input.internal_size2()) >> 1));
         }
 
         template<class SCALARTYPE, unsigned int ALIGNMENT>
         void transpose(viennacl::matrix<SCALARTYPE, viennacl::row_major, ALIGNMENT> const & input,
                        viennacl::matrix<SCALARTYPE, viennacl::row_major, ALIGNMENT> & output)
         {
-          viennacl::linalg::kernels::fft<SCALARTYPE, 1>::init();
-          
-            viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                             .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                             .get_kernel("transpose");
-            viennacl::ocl::enqueue(kernel(input,
-                                          output,
-                                          static_cast<cl_uint>(input.internal_size1()),
-                                          static_cast<cl_uint>(input.internal_size2() >> 1))
-                                  );
-        }
-        
-        template<class SCALARTYPE, unsigned int ALIGNMENT>
-        void real_to_complex(viennacl::vector<SCALARTYPE, ALIGNMENT> const & in,
-                             viennacl::vector<SCALARTYPE, ALIGNMENT> & out,
-                             std::size_t size) 
-        {
-          viennacl::linalg::kernels::fft<SCALARTYPE, 1>::init();
-            viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                             .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                             .get_kernel("real_to_complex");
-            viennacl::ocl::enqueue(kernel(in, out, static_cast<cl_uint>(size)));
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(viennacl::traits::opencl_handle(input).context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
+
+          viennacl::ocl::kernel& kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "transpose");
+          viennacl::ocl::enqueue(kernel(input,
+                                        output,
+                                        static_cast<cl_uint>(input.internal_size1()),
+                                        static_cast<cl_uint>(input.internal_size2() >> 1))
+                                );
         }
 
-        template<class SCALARTYPE, unsigned int ALIGNMENT>
-        void complex_to_real(viennacl::vector<SCALARTYPE, ALIGNMENT> const & in,
-                             viennacl::vector<SCALARTYPE, ALIGNMENT>& out,
-                             std::size_t size)
+        template<class SCALARTYPE>
+        void real_to_complex(viennacl::vector_base<SCALARTYPE> const & in,
+                             viennacl::vector_base<SCALARTYPE> & out,
+                             vcl_size_t size)
         {
-          viennacl::linalg::kernels::fft<SCALARTYPE, 1>::init();
-            viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                             .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                             .get_kernel("complex_to_real");
-            viennacl::ocl::enqueue(kernel(in, out, static_cast<cl_uint>(size)));
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(viennacl::traits::opencl_handle(in).context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
+          viennacl::ocl::kernel & kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "real_to_complex");
+          viennacl::ocl::enqueue(kernel(in, out, static_cast<cl_uint>(size)));
         }
 
-        template<class SCALARTYPE, unsigned int ALIGNMENT>
-        void reverse(viennacl::vector<SCALARTYPE, ALIGNMENT>& in)
+        template<class SCALARTYPE>
+        void complex_to_real(viennacl::vector_base<SCALARTYPE> const & in,
+                             viennacl::vector_base<SCALARTYPE>& out,
+                             vcl_size_t size)
         {
-          viennacl::linalg::kernels::fft<SCALARTYPE, 1>::init();
-            std::size_t size = in.size();
-            viennacl::ocl::kernel& kernel = viennacl::ocl::current_context()
-                                             .get_program(viennacl::linalg::kernels::fft<SCALARTYPE, 1>::program_name())
-                                             .get_kernel("reverse_inplace");
-            viennacl::ocl::enqueue(kernel(in, static_cast<cl_uint>(size)));
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(viennacl::traits::opencl_handle(in).context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
+          viennacl::ocl::kernel& kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "complex_to_real");
+          viennacl::ocl::enqueue(kernel(in, out, static_cast<cl_uint>(size)));
         }
 
-        
+        template<class SCALARTYPE>
+        void reverse(viennacl::vector_base<SCALARTYPE>& in)
+        {
+          viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(viennacl::traits::opencl_handle(in).context());
+          viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::init(ctx);
+          vcl_size_t size = in.size();
+          viennacl::ocl::kernel& kernel = ctx.get_kernel(viennacl::linalg::opencl::kernels::fft<SCALARTYPE>::program_name(), "reverse_inplace");
+          viennacl::ocl::enqueue(kernel(in, static_cast<cl_uint>(size)));
+        }
+
+
     } //namespace fft
   } //namespace detail
-  
+
   /**
     * @brief Generic inplace version of 1-D Fourier transformation.
     *
@@ -374,24 +370,24 @@ namespace viennacl {
     */
   template<class SCALARTYPE, unsigned int ALIGNMENT>
   void inplace_fft(viennacl::vector<SCALARTYPE, ALIGNMENT>& input,
-            std::size_t batch_num = 1,
+            vcl_size_t batch_num = 1,
             SCALARTYPE sign = -1.0)
   {
-      std::size_t size = (input.size() >> 1) / batch_num;
+      vcl_size_t size = (input.size() >> 1) / batch_num;
 
-      if(!detail::fft::is_radix2(size)) 
+      if(!viennacl::detail::fft::is_radix2(size))
       {
           viennacl::vector<SCALARTYPE, ALIGNMENT> output(input.size());
-          detail::fft::direct(input.handle(),
-                              output.handle(),
-                              size,
-                              size,
-                              batch_num,
-                              sign);
+          viennacl::detail::fft::direct(viennacl::traits::opencl_handle(input),
+                                        viennacl::traits::opencl_handle(output),
+                                        size,
+                                        size,
+                                        batch_num,
+                                        sign);
 
           viennacl::copy(output, input);
       } else {
-          detail::fft::radix2(input.handle(), size, size, batch_num, sign);
+          viennacl::detail::fft::radix2(viennacl::traits::opencl_handle(input), size, size, batch_num, sign);
       }
   }
 
@@ -406,23 +402,23 @@ namespace viennacl {
   template<class SCALARTYPE, unsigned int ALIGNMENT>
   void fft(viennacl::vector<SCALARTYPE, ALIGNMENT>& input,
             viennacl::vector<SCALARTYPE, ALIGNMENT>& output,
-            std::size_t batch_num = 1,
+            vcl_size_t batch_num = 1,
             SCALARTYPE sign = -1.0
             )
   {
-      std::size_t size = (input.size() >> 1) / batch_num;
+      vcl_size_t size = (input.size() >> 1) / batch_num;
 
-      if(detail::fft::is_radix2(size))
+      if(viennacl::detail::fft::is_radix2(size))
       {
           viennacl::copy(input, output);
-          detail::fft::radix2(output.handle(), size, size, batch_num, sign);
+          viennacl::detail::fft::radix2(viennacl::traits::opencl_handle(output), size, size, batch_num, sign);
       } else {
-          detail::fft::direct(input.handle(),
-                              output.handle(),
-                              size,
-                              size,
-                              batch_num,
-                              sign);
+          viennacl::detail::fft::direct(viennacl::traits::opencl_handle(input),
+                                        viennacl::traits::opencl_handle(output),
+                                        size,
+                                        size,
+                                        batch_num,
+                                        sign);
       }
   }
 
@@ -436,45 +432,45 @@ namespace viennacl {
   void inplace_fft(viennacl::matrix<SCALARTYPE, viennacl::row_major, ALIGNMENT>& input,
             SCALARTYPE sign = -1.0)
   {
-      std::size_t rows_num = input.size1();
-      std::size_t cols_num = input.size2() >> 1;
+      vcl_size_t rows_num = input.size1();
+      vcl_size_t cols_num = input.size2() >> 1;
 
-      std::size_t cols_int = input.internal_size2() >> 1;
+      vcl_size_t cols_int = input.internal_size2() >> 1;
 
       // batch with rows
-      if(detail::fft::is_radix2(cols_num)) 
+      if(viennacl::detail::fft::is_radix2(cols_num))
       {
-          detail::fft::radix2(input.handle(), cols_num, cols_int, rows_num, sign, detail::fft::FFT_DATA_ORDER::ROW_MAJOR);
-      } 
+          viennacl::detail::fft::radix2(viennacl::traits::opencl_handle(input), cols_num, cols_int, rows_num, sign, viennacl::detail::fft::FFT_DATA_ORDER::ROW_MAJOR);
+      }
       else
       {
           viennacl::matrix<SCALARTYPE, viennacl::row_major, ALIGNMENT> output(input.size1(), input.size2());
 
-          detail::fft::direct(input.handle(),
-                              output.handle(),
-                              cols_num,
-                              cols_int,
-                              rows_num,
-                              sign,
-                              detail::fft::FFT_DATA_ORDER::ROW_MAJOR
-                              );
+          viennacl::detail::fft::direct(viennacl::traits::opencl_handle(input),
+                                        viennacl::traits::opencl_handle(output),
+                                        cols_num,
+                                        cols_int,
+                                        rows_num,
+                                        sign,
+                                        viennacl::detail::fft::FFT_DATA_ORDER::ROW_MAJOR
+                                        );
 
           input = output;
       }
 
       // batch with cols
-      if (detail::fft::is_radix2(rows_num)) {
-          detail::fft::radix2(input.handle(), rows_num, cols_int, cols_num, sign, detail::fft::FFT_DATA_ORDER::COL_MAJOR);
+      if (viennacl::detail::fft::is_radix2(rows_num)) {
+          viennacl::detail::fft::radix2(viennacl::traits::opencl_handle(input), rows_num, cols_int, cols_num, sign, viennacl::detail::fft::FFT_DATA_ORDER::COL_MAJOR);
       } else {
           viennacl::matrix<SCALARTYPE, viennacl::row_major, ALIGNMENT> output(input.size1(), input.size2());
 
-          detail::fft::direct(input.handle(),
-                              output.handle(),
-                              rows_num,
-                              cols_int,
-                              cols_num,
-                              sign,
-                              detail::fft::FFT_DATA_ORDER::COL_MAJOR);
+          viennacl::detail::fft::direct(viennacl::traits::opencl_handle(input),
+                                        viennacl::traits::opencl_handle(output),
+                                        rows_num,
+                                        cols_int,
+                                        cols_num,
+                                        sign,
+                                        viennacl::detail::fft::FFT_DATA_ORDER::COL_MAJOR);
 
           input = output;
       }
@@ -493,46 +489,46 @@ namespace viennacl {
             viennacl::matrix<SCALARTYPE, viennacl::row_major, ALIGNMENT>& output,
             SCALARTYPE sign = -1.0)
   {
-      std::size_t rows_num = input.size1();
-      std::size_t cols_num = input.size2() >> 1;
+      vcl_size_t rows_num = input.size1();
+      vcl_size_t cols_num = input.size2() >> 1;
 
-      std::size_t cols_int = input.internal_size2() >> 1;
+      vcl_size_t cols_int = input.internal_size2() >> 1;
 
       // batch with rows
-      if(detail::fft::is_radix2(cols_num))
+      if(viennacl::detail::fft::is_radix2(cols_num))
       {
           output = input;
-          detail::fft::radix2(output.handle(), cols_num, cols_int, rows_num, sign, detail::fft::FFT_DATA_ORDER::ROW_MAJOR);
-      } 
+          viennacl::detail::fft::radix2(viennacl::traits::opencl_handle(output), cols_num, cols_int, rows_num, sign, viennacl::detail::fft::FFT_DATA_ORDER::ROW_MAJOR);
+      }
       else
       {
-          detail::fft::direct(input.handle(),
-                              output.handle(),
-                              cols_num,
-                              cols_int,
-                              rows_num,
-                              sign,
-                              detail::fft::FFT_DATA_ORDER::ROW_MAJOR
-                              );
+          viennacl::detail::fft::direct(viennacl::traits::opencl_handle(input),
+                                        viennacl::traits::opencl_handle(output),
+                                        cols_num,
+                                        cols_int,
+                                        rows_num,
+                                        sign,
+                                        viennacl::detail::fft::FFT_DATA_ORDER::ROW_MAJOR
+                                        );
       }
 
       // batch with cols
-      if(detail::fft::is_radix2(rows_num))
+      if(viennacl::detail::fft::is_radix2(rows_num))
       {
-          detail::fft::radix2(output.handle(), rows_num, cols_int, cols_num, sign, detail::fft::FFT_DATA_ORDER::COL_MAJOR);
-      } 
+          viennacl::detail::fft::radix2(viennacl::traits::opencl_handle(output), rows_num, cols_int, cols_num, sign, viennacl::detail::fft::FFT_DATA_ORDER::COL_MAJOR);
+      }
       else
       {
           viennacl::matrix<SCALARTYPE, viennacl::row_major, ALIGNMENT> tmp(output.size1(), output.size2());
           tmp = output;
 
-          detail::fft::direct(tmp.handle(),
-                              output.handle(),
+          viennacl::detail::fft::direct(viennacl::traits::opencl_handle(tmp),
+                              viennacl::traits::opencl_handle(output),
                               rows_num,
                               cols_int,
                               cols_num,
                               sign,
-                              detail::fft::FFT_DATA_ORDER::COL_MAJOR);
+                              viennacl::detail::fft::FFT_DATA_ORDER::COL_MAJOR);
       }
   }
 
@@ -547,10 +543,10 @@ namespace viennacl {
     */
   template<class SCALARTYPE, unsigned int ALIGNMENT>
   void inplace_ifft(viennacl::vector<SCALARTYPE, ALIGNMENT>& input,
-            std::size_t batch_num = 1)
+            vcl_size_t batch_num = 1)
   {
       viennacl::inplace_fft(input, batch_num, SCALARTYPE(1.0));
-      detail::fft::normalize(input);
+      viennacl::detail::fft::normalize(input);
   }
 
   /**
@@ -566,11 +562,11 @@ namespace viennacl {
   template<class SCALARTYPE, unsigned int ALIGNMENT>
   void ifft(viennacl::vector<SCALARTYPE, ALIGNMENT>& input,
             viennacl::vector<SCALARTYPE, ALIGNMENT>& output,
-            std::size_t batch_num = 1
+            vcl_size_t batch_num = 1
             )
   {
       viennacl::fft(input, output, batch_num, SCALARTYPE(1.0));
-      detail::fft::normalize(output);
+      viennacl::detail::fft::normalize(output);
   }
 
   namespace linalg
