@@ -2,24 +2,25 @@
 #define VIENNACL_LINALG_AMG_HPP_
 
 /* =========================================================================
-   Copyright (c) 2010-2011, Institute for Microelectronics,
+   Copyright (c) 2010-2014, Institute for Microelectronics,
                             Institute for Analysis and Scientific Computing,
                             TU Wien.
+   Portions of this software are copyright by UChicago Argonne, LLC.
 
                             -----------------
                   ViennaCL - The Vienna Computing Library
                             -----------------
 
    Project Head:    Karl Rupp                   rupp@iue.tuwien.ac.at
-               
+
    (A list of authors and contributors can be found in the PDF manual)
 
    License:         MIT (X11), see file LICENSE in the base directory
 ============================================================================= */
 
 /** @file viennacl/linalg/amg.hpp
-    @brief Main include file for algebraic multigrid (AMG) preconditioners.  Experimental in 1.2.x.
-    
+    @brief Main include file for algebraic multigrid (AMG) preconditioners.  Experimental.
+
     Implementation contributed by Markus Wagner
 */
 
@@ -29,7 +30,7 @@
 #include <boost/numeric/ublas/vector_proxy.hpp>
 #include <boost/numeric/ublas/matrix_proxy.hpp>
 #include <boost/numeric/ublas/vector.hpp>
-#include <boost/numeric/ublas/triangular.hpp> 
+#include <boost/numeric/ublas/triangular.hpp>
 #include <vector>
 #include <cmath>
 #include "viennacl/forwards.h"
@@ -43,7 +44,7 @@
 
 #include <map>
 
-#ifdef _OPENMP
+#ifdef VIENNACL_WITH_OPENMP
  #include <omp.h>
 #endif
 
@@ -55,45 +56,41 @@
 namespace viennacl
 {
   namespace linalg
-  {    
+  {
     typedef detail::amg::amg_tag          amg_tag;
-    
-    
-    
+
+
+
     /** @brief Setup AMG preconditioner
     *
     * @param A      Operator matrices on all levels
     * @param P      Prolongation/Interpolation operators on all levels
     * @param Pointvector  Vector of points on all levels
-    * @param tag    AMG preconditioner tag 
+    * @param tag    AMG preconditioner tag
     */
     template <typename InternalType1, typename InternalType2>
     void amg_setup(InternalType1 & A, InternalType1 & P, InternalType2 & Pointvector, amg_tag & tag)
     {
-      typedef typename InternalType1::value_type SparseMatrixType;
-      typedef typename InternalType2::value_type PointVectorType;     
-      typedef typename SparseMatrixType::value_type ScalarType;   
-      typedef typename SparseMatrixType::iterator1 InternalRowIterator;
-      typedef typename SparseMatrixType::iterator2 InternalColIterator;
-      
+      typedef typename InternalType2::value_type PointVectorType;
+
       unsigned int i, iterations, c_points, f_points;
       detail::amg::amg_slicing<InternalType1,InternalType2> Slicing;
-      
+
       // Set number of iterations. If automatic coarse grid construction is chosen (0), then set a maximum size and stop during the process.
       iterations = tag.get_coarselevels();
       if (iterations == 0)
         iterations = VIENNACL_AMG_MAX_LEVELS;
-             
+
       // For parallel coarsenings build data structures (number of threads set automatically).
       if (tag.get_coarse() == VIENNACL_AMG_COARSE_RS0 || tag.get_coarse() == VIENNACL_AMG_COARSE_RS3)
         Slicing.init(iterations);
-      
+
       for (i=0; i<iterations; ++i)
-      {  
+      {
         // Initialize Pointvector on level i and construct points.
-        Pointvector[i] = PointVectorType(A[i].size1());
+        Pointvector[i] = PointVectorType(static_cast<unsigned int>(A[i].size1()));
         Pointvector[i].init_points();
-        
+
         // Construct C and F points on coarse level (i is fine level, i+1 coarse level).
         detail::amg::amg_coarse (i, A, Pointvector, Slicing, tag);
 
@@ -101,32 +98,32 @@ namespace viennacl
         c_points = Pointvector[i].get_cpoints();
         f_points = Pointvector[i].get_fpoints();
 
-        #if defined (DEBUG) //or defined(DEBUGBENCH) 
+        #if defined (VIENNACL_AMG_DEBUG) //or defined(VIENNACL_AMG_DEBUGBENCH)
         std::cout << "Level " << i << ": ";
         std::cout << "No of C points = " << c_points << ", ";
         std::cout << "No of F points = " << f_points << std::endl;
         #endif
-        
+
         // Stop routine when the maximal coarse level is found (no C or F point). Coarsest level is level i.
         if (c_points == 0 || f_points == 0)
           break;
-          
+
         // Construct interpolation matrix for level i.
         detail::amg::amg_interpol (i, A, P, Pointvector, tag);
-        
+
         // Compute coarse grid operator (A[i+1] = R * A[i] * P) with R = trans(P).
         detail::amg::amg_galerkin_prod(A[i], P[i], A[i+1]);
-        
+
         // Test triple matrix product. Very slow for large matrix sizes (ublas).
         // test_triplematprod(A[i],P[i],A[i+1]);
-        
+
         Pointvector[i].delete_points();
-        
-        #ifdef DEBUG
+
+        #ifdef VIENNACL_AMG_DEBUG
         std::cout << "Coarse Grid Operator Matrix:" << std::endl;
         printmatrix (A[i+1]);
-        #endif  
-        
+        #endif
+
         // If Limit of coarse points is reached then stop. Coarsest level is level i+1.
         if (tag.get_coarselevels() == 0 && c_points <= VIENNACL_AMG_COARSE_LIMIT)
         {
@@ -134,23 +131,23 @@ namespace viennacl
           return;
         }
       }
-      tag.set_coarselevels(i);  
+      tag.set_coarselevels(i);
     }
-    
+
     /** @brief Initialize AMG preconditioner
     *
     * @param mat    System matrix
     * @param A      Operator matrices on all levels
     * @param P      Prolongation/Interpolation operators on all levels
     * @param Pointvector  Vector of points on all levels
-    * @param tag    AMG preconditioner tag 
+    * @param tag    AMG preconditioner tag
     */
     template <typename MatrixType, typename InternalType1, typename InternalType2>
     void amg_init(MatrixType const & mat, InternalType1 & A, InternalType1 & P, InternalType2 & Pointvector, amg_tag & tag)
     {
-      typedef typename MatrixType::value_type ScalarType;
+      //typedef typename MatrixType::value_type ScalarType;
       typedef typename InternalType1::value_type SparseMatrixType;
-      
+
       if (tag.get_coarselevels() > 0)
       {
         A.resize(tag.get_coarselevels()+1);
@@ -163,12 +160,12 @@ namespace viennacl
         P.resize(VIENNACL_AMG_MAX_LEVELS);
         Pointvector.resize(VIENNACL_AMG_MAX_LEVELS);
       }
-      
+
       // Insert operator matrix as operator for finest level.
       SparseMatrixType A0 (mat);
-      A.insert_element (0, A0);  
+      A.insert_element (0, A0);
     }
-    
+
     /** @brief Save operators after setup phase for CPU computation.
     *
     * @param A      Operator matrices on all levels on the CPU
@@ -176,18 +173,18 @@ namespace viennacl
     * @param R      Restriction operators on all levels on the CPU
     * @param A_setup    Operators matrices on all levels from setup phase
     * @param P_setup    Prolongation/Interpolation operators on all levels from setup phase
-    * @param tag    AMG preconditioner tag 
+    * @param tag    AMG preconditioner tag
     */
     template <typename InternalType1, typename InternalType2>
     void amg_transform_cpu (InternalType1 & A, InternalType1 & P, InternalType1 & R, InternalType2 & A_setup, InternalType2 & P_setup, amg_tag & tag)
-    {  
-      typedef typename InternalType1::value_type MatrixType;
-      
+    {
+      //typedef typename InternalType1::value_type MatrixType;
+
       // Resize internal data structures to actual size.
       A.resize(tag.get_coarselevels()+1);
       P.resize(tag.get_coarselevels());
       R.resize(tag.get_coarselevels());
-      
+
       // Transform into matrix type.
       for (unsigned int i=0; i<tag.get_coarselevels()+1; ++i)
       {
@@ -207,7 +204,7 @@ namespace viennacl
         P_setup[i].set_trans(false);
       }
     }
-    
+
     /** @brief Save operators after setup phase for GPU computation.
     *
     * @param A      Operator matrices on all levels on the GPU
@@ -215,57 +212,58 @@ namespace viennacl
     * @param R      Restriction operators on all levels on the GPU
     * @param A_setup    Operators matrices on all levels from setup phase
     * @param P_setup    Prolongation/Interpolation operators on all levels from setup phase
-    * @param tag    AMG preconditioner tag 
+    * @param tag    AMG preconditioner tag
+    * @param ctx      Optional context in which the auxiliary objects are created (one out of multiple OpenCL contexts, CUDA, host)
     */
     template <typename InternalType1, typename InternalType2>
-    void amg_transform_gpu (InternalType1 & A, InternalType1 & P, InternalType1 & R, InternalType2 & A_setup, InternalType2 & P_setup, amg_tag & tag)
+    void amg_transform_gpu (InternalType1 & A, InternalType1 & P, InternalType1 & R, InternalType2 & A_setup, InternalType2 & P_setup, amg_tag & tag, viennacl::context ctx)
     {
-      typedef typename InternalType1::value_type MatrixType;
-      typedef typename InternalType2::value_type::value_type ScalarType;
-      
       // Resize internal data structures to actual size.
       A.resize(tag.get_coarselevels()+1);
       P.resize(tag.get_coarselevels());
       R.resize(tag.get_coarselevels());
-      
+
       // Copy to GPU using the internal sparse matrix structure: std::vector<std::map>.
       for (unsigned int i=0; i<tag.get_coarselevels()+1; ++i)
       {
-        A[i].resize(A_setup[i].size1(),A_setup[i].size2(),false);
+        viennacl::switch_memory_context(A[i], ctx);
+        //A[i].resize(A_setup[i].size1(),A_setup[i].size2(),false);
         viennacl::copy(*(A_setup[i].get_internal_pointer()),A[i]);
       }
       for (unsigned int i=0; i<tag.get_coarselevels(); ++i)
       {
-        P[i].resize(P_setup[i].size1(),P_setup[i].size2(),false);
+        viennacl::switch_memory_context(P[i], ctx);
+        //P[i].resize(P_setup[i].size1(),P_setup[i].size2(),false);
         viennacl::copy(*(P_setup[i].get_internal_pointer()),P[i]);
         //viennacl::copy((boost::numeric::ublas::compressed_matrix<ScalarType>)P_setup[i],P[i]);
       }
       for (unsigned int i=0; i<tag.get_coarselevels(); ++i)
       {
-        R[i].resize(P_setup[i].size2(),P_setup[i].size1(),false);
+        viennacl::switch_memory_context(R[i], ctx);
+        //R[i].resize(P_setup[i].size2(),P_setup[i].size1(),false);
         P_setup[i].set_trans(true);
         viennacl::copy(*(P_setup[i].get_internal_pointer()),R[i]);
         P_setup[i].set_trans(false);
       }
     }
-    
+
     /** @brief Setup data structures for precondition phase.
     *
     * @param result    Result vector on all levels
     * @param rhs    RHS vector on all levels
     * @param residual    Residual vector on all levels
     * @param A      Operators matrices on all levels from setup phase
-    * @param tag    AMG preconditioner tag 
+    * @param tag    AMG preconditioner tag
     */
     template <typename InternalVectorType, typename SparseMatrixType>
     void amg_setup_apply (InternalVectorType & result, InternalVectorType & rhs, InternalVectorType & residual, SparseMatrixType const & A, amg_tag const & tag)
-    {        
+    {
       typedef typename InternalVectorType::value_type VectorType;
-      
+
       result.resize(tag.get_coarselevels()+1);
       rhs.resize(tag.get_coarselevels()+1);
       residual.resize(tag.get_coarselevels());
-            
+
       for (unsigned int level=0; level < tag.get_coarselevels()+1; ++level)
       {
         result[level] = VectorType(A[level].size1());
@@ -279,6 +277,38 @@ namespace viennacl
         residual[level].clear();
       }
     }
+
+
+    /** @brief Setup data structures for precondition phase for later use on the GPU
+    *
+    * @param result    Result vector on all levels
+    * @param rhs    RHS vector on all levels
+    * @param residual    Residual vector on all levels
+    * @param A      Operators matrices on all levels from setup phase
+    * @param tag    AMG preconditioner tag
+    * @param ctx      Optional context in which the auxiliary objects are created (one out of multiple OpenCL contexts, CUDA, host)
+    */
+    template <typename InternalVectorType, typename SparseMatrixType>
+    void amg_setup_apply (InternalVectorType & result, InternalVectorType & rhs, InternalVectorType & residual, SparseMatrixType const & A, amg_tag const & tag, viennacl::context ctx)
+    {
+      typedef typename InternalVectorType::value_type VectorType;
+
+      result.resize(tag.get_coarselevels()+1);
+      rhs.resize(tag.get_coarselevels()+1);
+      residual.resize(tag.get_coarselevels());
+
+      for (unsigned int level=0; level < tag.get_coarselevels()+1; ++level)
+      {
+        result[level] = VectorType(A[level].size1(), ctx);
+        rhs[level] = VectorType(A[level].size1(), ctx);
+      }
+      for (unsigned int level=0; level < tag.get_coarselevels(); ++level)
+      {
+        residual[level] = VectorType(A[level].size1(), ctx);
+      }
+    }
+
+
     /** @brief Pre-compute LU factorization for direct solve (ublas library).
      *  @brief Speeds up precondition phase as this is computed only once overall instead of once per iteration.
     *
@@ -287,22 +317,22 @@ namespace viennacl
     * @param A      Operator matrix on coarsest level
     */
     template <typename ScalarType, typename SparseMatrixType>
-    void amg_lu(boost::numeric::ublas::compressed_matrix<ScalarType> & op, boost::numeric::ublas::permutation_matrix<ScalarType> & Permutation, SparseMatrixType const & A)
+    void amg_lu(boost::numeric::ublas::compressed_matrix<ScalarType> & op, boost::numeric::ublas::permutation_matrix<> & Permutation, SparseMatrixType const & A)
     {
       typedef typename SparseMatrixType::const_iterator1 ConstRowIterator;
       typedef typename SparseMatrixType::const_iterator2 ConstColIterator;
-    
-      // Copy to operator matrix. Needed 
+
+      // Copy to operator matrix. Needed
       op.resize(A.size1(),A.size2(),false);
       for (ConstRowIterator row_iter = A.begin1(); row_iter != A.end1(); ++row_iter)
         for (ConstColIterator col_iter = row_iter.begin(); col_iter != row_iter.end(); ++col_iter)
           op (col_iter.index1(), col_iter.index2()) = *col_iter;
-      
+
       // Permutation matrix has to be reinitialized with actual size. Do not clear() or resize()!
-      Permutation = boost::numeric::ublas::permutation_matrix<ScalarType> (op.size1());
+      Permutation = boost::numeric::ublas::permutation_matrix<> (op.size1());
       boost::numeric::ublas::lu_factorize(op,Permutation);
     }
-    
+
     /** @brief AMG preconditioner class, can be supplied to solve()-routines
     */
     template <typename MatrixType>
@@ -312,31 +342,31 @@ namespace viennacl
       typedef boost::numeric::ublas::vector<ScalarType> VectorType;
       typedef detail::amg::amg_sparsematrix<ScalarType> SparseMatrixType;
       typedef detail::amg::amg_pointvector PointVectorType;
-      
+
       typedef typename SparseMatrixType::const_iterator1 InternalConstRowIterator;
       typedef typename SparseMatrixType::const_iterator2 InternalConstColIterator;
       typedef typename SparseMatrixType::iterator1 InternalRowIterator;
       typedef typename SparseMatrixType::iterator2 InternalColIterator;
-      
+
       boost::numeric::ublas::vector <SparseMatrixType> A_setup;
       boost::numeric::ublas::vector <SparseMatrixType> P_setup;
       boost::numeric::ublas::vector <MatrixType> A;
       boost::numeric::ublas::vector <MatrixType> P;
       boost::numeric::ublas::vector <MatrixType> R;
       boost::numeric::ublas::vector <PointVectorType> Pointvector;
-       
+
       mutable boost::numeric::ublas::compressed_matrix<ScalarType> op;
-      mutable boost::numeric::ublas::permutation_matrix<ScalarType> Permutation;  
-      
+      mutable boost::numeric::ublas::permutation_matrix<> Permutation;
+
       mutable boost::numeric::ublas::vector <VectorType> result;
       mutable boost::numeric::ublas::vector <VectorType> rhs;
       mutable boost::numeric::ublas::vector <VectorType> residual;
-      
+
       mutable bool done_init_apply;
-          
-      amg_tag _tag;
+
+      amg_tag tag_;
     public:
-    
+
       amg_precond(): Permutation(0) {}
       /** @brief The constructor. Saves system matrix, tag and builds data structures for setup.
       *
@@ -345,25 +375,25 @@ namespace viennacl
       */
       amg_precond(MatrixType const & mat, amg_tag const & tag): Permutation(0)
       {
-        _tag = tag;
+        tag_ = tag;
         // Initialize data structures.
-        amg_init (mat,A_setup,P_setup,Pointvector,_tag);
-        
+        amg_init (mat,A_setup,P_setup,Pointvector,tag_);
+
         done_init_apply = false;
       }
-      
+
       /** @brief Start setup phase for this class and copy data structures.
       */
       void setup()
       {
         // Start setup phase.
-        amg_setup(A_setup,P_setup,Pointvector,_tag);
+        amg_setup(A_setup,P_setup,Pointvector,tag_);
         // Transform to CPU-Matrixtype for precondition phase.
-        amg_transform_cpu(A,P,R,A_setup,P_setup,_tag);
-        
+        amg_transform_cpu(A,P,R,A_setup,P_setup,tag_);
+
         done_init_apply = false;
       }
-      
+
       /** @brief Prepare data structures for preconditioning:
        *  Build data structures for precondition phase.
        *  Do LU factorization on coarsest level.
@@ -371,13 +401,13 @@ namespace viennacl
       void init_apply() const
       {
         // Setup precondition phase (Data structures).
-        amg_setup_apply(result,rhs,residual,A_setup,_tag);
+        amg_setup_apply(result,rhs,residual,A_setup,tag_);
         // Do LU factorization for direct solve.
-        amg_lu(op,Permutation,A_setup[_tag.get_coarselevels()]);
-        
+        amg_lu(op,Permutation,A_setup[tag_.get_coarselevels()]);
+
         done_init_apply = true;
       }
-      
+
       /** @brief Returns complexity measures.
       *
       * @param avgstencil  Average stencil sizes on all levels
@@ -386,14 +416,14 @@ namespace viennacl
       template <typename VectorType>
       ScalarType calc_complexity(VectorType & avgstencil)
       {
-        avgstencil = VectorType (_tag.get_coarselevels()+1);
+        avgstencil = VectorType (tag_.get_coarselevels()+1);
         unsigned int nonzero=0, systemmat_nonzero=0, level_coefficients=0;
-          
-        for (unsigned int level=0; level < _tag.get_coarselevels()+1; ++level)
+
+        for (unsigned int level=0; level < tag_.get_coarselevels()+1; ++level)
         {
           level_coefficients = 0;
           for (InternalRowIterator row_iter = A_setup[level].begin1(); row_iter != A_setup[level].end1(); ++row_iter)
-          { 
+          {
             for (InternalColIterator col_iter = row_iter.begin(); col_iter != row_iter.end(); ++col_iter)
             {
               if (level == 0)
@@ -402,9 +432,9 @@ namespace viennacl
               level_coefficients++;
             }
           }
-          avgstencil[level] = level_coefficients/(double)A_setup[level].size1();
+          avgstencil[level] = level_coefficients/static_cast<ScalarType>(A_setup[level].size1());
         }
-        return nonzero/static_cast<double>(systemmat_nonzero);
+        return nonzero/static_cast<ScalarType>(systemmat_nonzero);
       }
 
       /** @brief Precondition Operation
@@ -413,82 +443,82 @@ namespace viennacl
       */
       template <typename VectorType>
       void apply(VectorType & vec) const
-      {   
+      {
         // Build data structures and do lu factorization before first iteration step.
         if (!done_init_apply)
           init_apply();
-        
+
         int level;
-        
+
         // Precondition operation (Yang, p.3)
         rhs[0] = vec;
-        for (level=0; level <(signed)_tag.get_coarselevels(); level++)
-        {    
+        for (level=0; level <static_cast<int>(tag_.get_coarselevels()); level++)
+        {
           result[level].clear();
-          
-          // Apply Smoother _presmooth times.
-          smooth_jacobi (level, _tag.get_presmooth(), result[level], rhs[level]);    
-          
-          #ifdef DEBUG
+
+          // Apply Smoother presmooth_ times.
+          smooth_jacobi (level, tag_.get_presmooth(), result[level], rhs[level]);
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "After presmooth:" << std::endl;
           printvector(result[level]);
           #endif
 
           // Compute residual.
           residual[level] = rhs[level] - boost::numeric::ublas::prod (A[level],result[level]);
-          
-          #ifdef DEBUG          
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "Residual:" << std::endl;
           printvector(residual[level]);
           #endif
-          
+
           // Restrict to coarse level. Restricted residual is RHS of coarse level.
           rhs[level+1] = boost::numeric::ublas::prod (R[level],residual[level]);
-          
-          #ifdef DEBUG
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "Restricted Residual: " << std::endl;
           printvector(rhs[level+1]);
           #endif
         }
-          
+
         // On highest level use direct solve to solve equation.
         result[level] = rhs[level];
         boost::numeric::ublas::lu_substitute(op,Permutation,result[level]);
 
-        #ifdef DEBUG
+        #ifdef VIENNACL_AMG_DEBUG
         std::cout << "After direct solve: " << std::endl;
         printvector (result[level]);
         #endif
-          
-        for (level=_tag.get_coarselevels()-1; level >= 0; level--)
-        {       
-          #ifdef DEBUG
+
+        for (level=tag_.get_coarselevels()-1; level >= 0; level--)
+        {
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "Coarse Error: " << std::endl;
           printvector(result[level+1]);
           #endif
-          
+
           // Interpolate error to fine level. Correct solution by adding error.
           result[level] += boost::numeric::ublas::prod (P[level], result[level+1]);
-              
-          #ifdef DEBUG
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "Corrected Result: " << std::endl;
           printvector (result[level]);
           #endif
-          
-          // Apply Smoother _postsmooth times.
-          smooth_jacobi (level, _tag.get_postsmooth(), result[level], rhs[level]);
-          
-          #ifdef DEBUG
+
+          // Apply Smoother postsmooth_ times.
+          smooth_jacobi (level, tag_.get_postsmooth(), result[level], rhs[level]);
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "After postsmooth: " << std::endl;
           printvector (result[level]);
           #endif
-        }    
+        }
         vec = result[0];
       }
-      
+
       /** @brief (Weighted) Jacobi Smoother (CPU version)
       * @param level    Coarse level to which smoother is applied to
-      * @param iterations  Number of smoother iterations    
+      * @param iterations  Number of smoother iterations
       * @param x     The vector smoothing is applied to
       * @param rhs    The right hand side of the equation for the smoother
       */
@@ -496,20 +526,20 @@ namespace viennacl
       void smooth_jacobi(int level, int const iterations, VectorType & x, VectorType const & rhs) const
       {
         VectorType old_result (x.size());
-        unsigned int index;
+        long index;
         ScalarType sum = 0, diag = 1;
-        
+
         for (int i=0; i<iterations; ++i)
         {
           old_result = x;
           x.clear();
-#ifdef _OPENMP
+#ifdef VIENNACL_WITH_OPENMP
           #pragma omp parallel for private (sum,diag) shared (rhs,x)
-#endif          
-          for (index=0; index<A_setup[level].size1(); ++index)  
+#endif
+          for (index=0; index < static_cast<long>(A_setup[level].size1()); ++index)
           {
             InternalConstRowIterator row_iter = A_setup[level].begin1();
-            row_iter += index; 
+            row_iter += index;
             sum = 0;
             diag = 1;
             for (InternalConstColIterator col_iter = row_iter.begin(); col_iter != row_iter.end(); ++col_iter)
@@ -519,14 +549,14 @@ namespace viennacl
               else
                 sum += *col_iter * old_result[col_iter.index2()];
             }
-            x[index]= _tag.get_jacobiweight() * (rhs[index] - sum) / diag + (1-_tag.get_jacobiweight()) * old_result[index];
+            x[index]= static_cast<ScalarType>(tag_.get_jacobiweight()) * (rhs[index] - sum) / diag + (1-static_cast<ScalarType>(tag_.get_jacobiweight())) * old_result[index];
           }
         }
       }
-      
-      amg_tag & tag() { return _tag; }
+
+      amg_tag & tag() { return tag_; }
     };
-    
+
     /** @brief AMG preconditioner class, can be supplied to solve()-routines.
     *
     *  Specialization for compressed_matrix
@@ -538,65 +568,67 @@ namespace viennacl
       typedef viennacl::vector<ScalarType> VectorType;
       typedef detail::amg::amg_sparsematrix<ScalarType> SparseMatrixType;
       typedef detail::amg::amg_pointvector PointVectorType;
-      
+
       typedef typename SparseMatrixType::const_iterator1 InternalConstRowIterator;
       typedef typename SparseMatrixType::const_iterator2 InternalConstColIterator;
       typedef typename SparseMatrixType::iterator1 InternalRowIterator;
       typedef typename SparseMatrixType::iterator2 InternalColIterator;
-      
+
       boost::numeric::ublas::vector <SparseMatrixType> A_setup;
       boost::numeric::ublas::vector <SparseMatrixType> P_setup;
       boost::numeric::ublas::vector <MatrixType> A;
       boost::numeric::ublas::vector <MatrixType> P;
       boost::numeric::ublas::vector <MatrixType> R;
       boost::numeric::ublas::vector <PointVectorType> Pointvector;
-      
+
       mutable boost::numeric::ublas::compressed_matrix<ScalarType> op;
-      mutable boost::numeric::ublas::permutation_matrix<ScalarType> Permutation;  
-      
+      mutable boost::numeric::ublas::permutation_matrix<> Permutation;
+
       mutable boost::numeric::ublas::vector <VectorType> result;
       mutable boost::numeric::ublas::vector <VectorType> rhs;
       mutable boost::numeric::ublas::vector <VectorType> residual;
-          
+
+      viennacl::context ctx_;
+
       mutable bool done_init_apply;
-    
-      amg_tag _tag;
-      
+
+      amg_tag tag_;
+
     public:
-      
+
       amg_precond(): Permutation(0) {}
-      
+
       /** @brief The constructor. Builds data structures.
       *
       * @param mat  System matrix
       * @param tag  The AMG tag
       */
-      amg_precond(compressed_matrix<ScalarType, MAT_ALIGNMENT> const & mat, amg_tag const & tag): Permutation(0)
+      amg_precond(compressed_matrix<ScalarType, MAT_ALIGNMENT> const & mat, amg_tag const & tag): Permutation(0), ctx_(viennacl::traits::context(mat))
       {
-        _tag = tag;
-        
+        tag_ = tag;
+
         // Copy to CPU. Internal structure of sparse matrix is used for copy operation.
         std::vector<std::map<unsigned int, ScalarType> > mat2 = std::vector<std::map<unsigned int, ScalarType> >(mat.size1());
         viennacl::copy(mat, mat2);
-        
+
         // Initialize data structures.
-        amg_init (mat2,A_setup,P_setup,Pointvector,_tag);
-          
+        amg_init (mat2,A_setup,P_setup,Pointvector,tag_);
+
         done_init_apply = false;
       }
-      
+
       /** @brief Start setup phase for this class and copy data structures.
       */
       void setup()
       {
         // Start setup phase.
-        amg_setup(A_setup,P_setup,Pointvector,_tag);  
+        amg_setup(A_setup,P_setup,Pointvector, tag_);
         // Transform to GPU-Matrixtype for precondition phase.
-        amg_transform_gpu(A,P,R,A_setup,P_setup,_tag);  
-        
+        amg_transform_gpu(A,P,R,A_setup,P_setup, tag_, ctx_);
+
         done_init_apply = false;
       }
-      
+
       /** @brief Prepare data structures for preconditioning:
        *  Build data structures for precondition phase.
        *  Do LU factorization on coarsest level.
@@ -604,10 +636,10 @@ namespace viennacl
       void init_apply() const
       {
         // Setup precondition phase (Data structures).
-        amg_setup_apply(result,rhs,residual,A_setup,_tag);
+        amg_setup_apply(result,rhs,residual,A_setup,tag_, ctx_);
         // Do LU factorization for direct solve.
-        amg_lu(op,Permutation,A_setup[_tag.get_coarselevels()]);
-        
+        amg_lu(op,Permutation,A_setup[tag_.get_coarselevels()]);
+
         done_init_apply = true;
       }
 
@@ -619,14 +651,14 @@ namespace viennacl
       template <typename VectorType>
       ScalarType calc_complexity(VectorType & avgstencil)
       {
-        avgstencil = VectorType (_tag.get_coarselevels()+1);
+        avgstencil = VectorType (tag_.get_coarselevels()+1);
         unsigned int nonzero=0, systemmat_nonzero=0, level_coefficients=0;
-        
-        for (unsigned int level=0; level < _tag.get_coarselevels()+1; ++level)
+
+        for (unsigned int level=0; level < tag_.get_coarselevels()+1; ++level)
         {
           level_coefficients = 0;
           for (InternalRowIterator row_iter = A_setup[level].begin1(); row_iter != A_setup[level].end1(); ++row_iter)
-          { 
+          {
             for (InternalColIterator col_iter = row_iter.begin(); col_iter != row_iter.end(); ++col_iter)
             {
               if (level == 0)
@@ -642,48 +674,50 @@ namespace viennacl
 
       /** @brief Precondition Operation
       *
-      * @param vec The vector to which preconditioning is applied to 
+      * @param vec The vector to which preconditioning is applied to
       */
       template <typename VectorType>
       void apply(VectorType & vec) const
       {
         if (!done_init_apply)
-          init_apply();  
-        
+          init_apply();
+
         int level;
-        
+
         // Precondition operation (Yang, p.3).
         rhs[0] = vec;
-        for (level=0; level <(signed)_tag.get_coarselevels(); level++)
-        {    
+        for (level=0; level <static_cast<int>(tag_.get_coarselevels()); level++)
+        {
           result[level].clear();
-          
-          // Apply Smoother _presmooth times.
-          smooth_jacobi (level, _tag.get_presmooth(), result[level], rhs[level]);
 
-          #ifdef DEBUG
+          // Apply Smoother presmooth_ times.
+          smooth_jacobi (level, tag_.get_presmooth(), result[level], rhs[level]);
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "After presmooth: " << std::endl;
           printvector(result[level]);
           #endif
-          
+
           // Compute residual.
-          residual[level] = rhs[level] - viennacl::linalg::prod (A[level],result[level]);
-          
-          #ifdef DEBUG             
+          //residual[level] = rhs[level] - viennacl::linalg::prod (A[level],result[level]);
+          residual[level] = viennacl::linalg::prod (A[level],result[level]);
+          residual[level] = rhs[level] - residual[level];
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "Residual: " << std::endl;
           printvector(residual[level]);
           #endif
-          
+
           // Restrict to coarse level. Result is RHS of coarse level equation.
           //residual_coarse[level] = viennacl::linalg::prod(R[level],residual[level]);
           rhs[level+1] = viennacl::linalg::prod(R[level],residual[level]);
-          
-          #ifdef DEBUG
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "Restricted Residual: " << std::endl;
           printvector(rhs[level+1]);
           #endif
         }
-          
+
         // On highest level use direct solve to solve equation (on the CPU)
         //TODO: Use GPU direct solve!
         result[level] = rhs[level];
@@ -692,71 +726,69 @@ namespace viennacl
         copy (result[level],result_cpu);
         boost::numeric::ublas::lu_substitute(op,Permutation,result_cpu);
         copy (result_cpu, result[level]);
-        
-        #ifdef DEBUG
+
+        #ifdef VIENNACL_AMG_DEBUG
         std::cout << "After direct solve: " << std::endl;
         printvector (result[level]);
         #endif
-          
-        for (level=_tag.get_coarselevels()-1; level >= 0; level--)
-        {   
-          #ifdef DEBUG
+
+        for (level=tag_.get_coarselevels()-1; level >= 0; level--)
+        {
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "Coarse Error: " << std::endl;
           printvector(result[level+1]);
           #endif
-          
+
           // Interpolate error to fine level and correct solution.
           result[level] += viennacl::linalg::prod(P[level],result[level+1]);
-              
-          #ifdef DEBUG
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "Corrected Result: " << std::endl;
           printvector (result[level]);
           #endif
-          
-          // Apply Smoother _postsmooth times.
-          smooth_jacobi (level, _tag.get_postsmooth(), result[level], rhs[level]);
-          
-          #ifdef DEBUG
+
+          // Apply Smoother postsmooth_ times.
+          smooth_jacobi (level, tag_.get_postsmooth(), result[level], rhs[level]);
+
+          #ifdef VIENNACL_AMG_DEBUG
           std::cout << "After postsmooth: " << std::endl;
           printvector (result[level]);
           #endif
-        }    
+        }
         vec = result[0];
       }
-      
+
       /** @brief Jacobi Smoother (GPU version)
       * @param level       Coarse level to which smoother is applied to
-      * @param iterations  Number of smoother iterations    
+      * @param iterations  Number of smoother iterations
       * @param x           The vector smoothing is applied to
       * @param rhs         The right hand side of the equation for the smoother
       */
       template <typename VectorType>
       void smooth_jacobi(int level, unsigned int iterations, VectorType & x, VectorType const & rhs) const
-      {     
-        VectorType old_result (x.size());
-  
-  //viennacl::ocl::program & p = viennacl::ocl::current_context().add_program
-  //          (viennacl::tools::make_double_kernel(jacobi_kernel,viennacl::ocl::current_device().info()), "jacobi_kernel");
-  //viennacl::ocl::kernel & k = p.add_kernel("jacobi");
-  
-        viennacl::ocl::kernel & k = viennacl::ocl::get_kernel(viennacl::linalg::kernels::compressed_matrix<ScalarType, MAT_ALIGNMENT>::program_name(),
-                    "jacobi");
-  
+      {
+        VectorType old_result = x;
+
+        viennacl::ocl::context & ctx = const_cast<viennacl::ocl::context &>(viennacl::traits::opencl_handle(x).context());
+        viennacl::linalg::opencl::kernels::compressed_matrix<ScalarType>::init(ctx);
+        viennacl::ocl::kernel & k = ctx.get_kernel(viennacl::linalg::opencl::kernels::compressed_matrix<ScalarType>::program_name(), "jacobi");
+
         for (unsigned int i=0; i<iterations; ++i)
         {
-          old_result = x;    
+          if (i > 0)
+            old_result = x;
           x.clear();
-          viennacl::ocl::enqueue(k(A[level].handle1(), A[level].handle2(), A[level].handle(),
-                                  static_cast<ScalarType>(_tag.get_jacobiweight()), 
-                                  old_result,
-                                  x,
-                                  rhs,
-                                  static_cast<cl_uint>(rhs.size()))); 
-          
+          viennacl::ocl::enqueue(k(A[level].handle1().opencl_handle(), A[level].handle2().opencl_handle(), A[level].handle().opencl_handle(),
+                                  static_cast<ScalarType>(tag_.get_jacobiweight()),
+                                  viennacl::traits::opencl_handle(old_result),
+                                  viennacl::traits::opencl_handle(x),
+                                  viennacl::traits::opencl_handle(rhs),
+                                  static_cast<cl_uint>(rhs.size())));
+
         }
       }
-      
-      amg_tag & tag() { return _tag; }
+
+      amg_tag & tag() { return tag_; }
     };
 
   }
